import 'package:rxdart/rxdart.dart';
import 'package:flutter/widgets.dart';

/// DataPoint<T> is a wrapper for RxDart `BehaviorSubject`.
/// ### Remember to dispose to avoid leaking streams!
///
/// DataPoints should probably be private to your store.
/// You should only expose `public`, which is read-only.
@immutable
class DataPoint<T> {
  DataPoint({T initial})
      : _subject = (initial != null)
            ? BehaviorSubject<T>.seeded(initial)
            : BehaviorSubject<T>();
  final BehaviorSubject<T> _subject;
  T get value => _subject.value;
  set value(T newValue) => _subject.add(newValue);
  Stream<T> get stream => _subject.stream;
  Future<dynamic> dispose() => _subject.close();
  PublicDataPoint<T> get public => PublicDataPoint<T>(this);
  String toString() => value.toString();
}

/// PublicDataPoint is a wrapper for `DataPoint` that is read-only.
/// The `use()` method allows you to build a Widget with the DataPoint.
/// It will then automatically rebuild when the data changes.
@immutable
class PublicDataPoint<T> {
  PublicDataPoint(this._mut);
  final DataPoint<T> _mut;
  Stream<T> get stream => _mut.stream;
  T get value => _mut.value;

  Widget use(Widget Function(BuildContext context, T data) builder) {
    return StreamBuilder<T>(
      initialData: value,
      stream: stream,
      builder: (BuildContext context, AsyncSnapshot<T> snapshot) {
        return builder(context, snapshot.data);
      },
    );
  }

  String toString() => value.toString();
}

/// Store<T> is an InheritedWidget that stores data.
///
/// If you need to do cleanup when the Widget is being disposed,
/// use a StoreManager<T> instead.
///
/// Usage example:
/// `Store<String>(child: MyWidget(), data: "A string");`
///
/// Then, when you want to get the data associated with the store:
/// `Store.of<String>(context);`
///
/// You can have multiple Stores in one widget tree, but you can only access the latest
/// Store of a given Type.
@immutable
class Store<T> extends InheritedWidget {
  Store({@required Widget child, @required this.data}) : super(child: child);
  final T data;
  @override
  bool updateShouldNotify(InheritedWidget old) => true;
  static Type _typeOf<T>() => T;
  static T of<T>(BuildContext context) =>
      (context.inheritFromWidgetOfExactType(_typeOf<Store<T>>()) as Store<T>)
          .data;
}

/// StoreManager contains a `Store` and calls `data.dispose()` when you dispose it.
/// This is useful when you want to clean up some stuff (like Streams)
@immutable
class StoreManager<T> extends StatefulWidget {
  StoreManager({@required this.child, @required this.data});
  final T data;
  final Widget child;
  @override
  _StoreManagerState<T> createState() => _StoreManagerState<T>();
}

class _StoreManagerState<T> extends State<StoreManager> {
  @override
  Widget build(BuildContext context) {
    return Store<T>(child: widget.child, data: widget.data);
  }

  void dispose() {
    if (widget.data.dispose != null) {
      widget.data.dispose();
    }
    super.dispose();
  }
}

abstract class StoreData {
  void dispose() {}
}
